==> cirq_unstable-0.7.0.dev20191121170219-py3-none-any.whl: cirq/ops/matrix_gates.py <==
# Copyright 2018 The Cirq Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Quantum gates defined by a matrix."""

from typing import cast, Any, Tuple, Optional, Iterable, TYPE_CHECKING

import numpy as np

from cirq import linalg, protocols
from cirq._compat import proper_repr, deprecated
from cirq.ops import gate_features, raw_types

if TYPE_CHECKING:
    import cirq


class MatrixGate(raw_types.Gate):
    """A unitary qubit or qudit gate defined entirely by its matrix."""

    def __init__(self,
                 matrix: np.ndarray,
                 *,
                 qid_shape: Optional[Iterable[int]] = None) -> None:
        """Initializes a matrix gate.
        Args:
            matrix: The matrix that defines the gate.
            qid_shape: The shape of state tensor that the matrix applies to.
                If not specified, this value is inferred by assuming that the
                matrix is supposed to apply to qubits.
        """
        if len(matrix.shape) != 2 or matrix.shape[0] != matrix.shape[1]:
            raise ValueError('`matrix` must be a square 2d numpy array.')

        if qid_shape is None:
            n = int(np.round(np.log2(matrix.shape[0] or 1)))
            if 2**n != matrix.shape[0]:
                raise ValueError(
                    f'Matrix width ({matrix.shape[0]}) is not a power of 2 and '
                    f'qid_shape is not specified.')
            qid_shape = (2,) * n

        self._matrix = matrix
        self._qid_shape = tuple(qid_shape)
        m = int(np.prod(self._qid_shape))
        if self._matrix.shape != (m, m):
            raise ValueError('Wrong matrix shape for qid_shape.\n'
                             f'Matrix shape: {self._matrix.shape}\n'
                             f'qid_shape: {self._qid_shape}\n')

        if not linalg.is_unitary(matrix):
            raise ValueError(f'Not a unitary matrix: {self._matrix}')

    def _json_dict_(self):
        return {
            'cirq_type': self.__class__.__name__,
            'matrix': self._matrix.tolist(),
            'qid_shape': self._qid_shape,
        }

    @classmethod
    def _from_json_dict_(cls, matrix, qid_shape, **kwargs):
        return cls(matrix=np.array(matrix), qid_shape=qid_shape)

    def _qid_shape_(self) -> Tuple[int, ...]:
        return self._qid_shape

    def __pow__(self, exponent: Any) -> 'MatrixGate':
        if not isinstance(exponent, (int, float)):
            return NotImplemented
        e = cast(float, exponent)
        new_mat = linalg.map_eigenvalues(self._matrix, lambda b: b**e)
        return MatrixGate(new_mat, qid_shape=self._qid_shape)

    def _phase_by_(self, phase_turns: float, qubit_index: int) -> 'MatrixGate':
        if not isinstance(phase_turns, (int, float)):
            return NotImplemented
        if self._qid_shape[qubit_index] != 2:
            return NotImplemented
        result = np.copy(self._matrix).reshape(self._qid_shape * 2)

        p = np.exp(2j * np.pi * phase_turns)
        i = qubit_index
        j = qubit_index + len(self._qid_shape)
        result[linalg.slice_for_qubits_equal_to([i], 1)] *= p
        result[linalg.slice_for_qubits_equal_to([j], 1)] *= np.conj(p)
        return MatrixGate(matrix=result.reshape(self._matrix.shape),
                          qid_shape=self._qid_shape)

    def _has_unitary_(self) -> bool:
        return True

    def _unitary_(self) -> np.ndarray:
        return np.copy(self._matrix)

    def _circuit_diagram_info_(self, args: 'cirq.CircuitDiagramInfoArgs'
                              ) -> 'cirq.CircuitDiagramInfo':
        main = _matrix_to_diagram_symbol(self._matrix, args)
        rest = [f'#{i+1}' for i in range(1, len(self._qid_shape))]
        return protocols.CircuitDiagramInfo(wire_symbols=[main, *rest])

    def __hash__(self):
        vals = tuple(v for _, v in np.ndenumerate(self._matrix))
        return hash((MatrixGate, vals))

    def _approx_eq_(self, other: Any, atol) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return np.allclose(self._matrix, other._matrix, rtol=0, atol=atol)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return (self._qid_shape == other._qid_shape and
                np.array_equal(self._matrix, other._matrix))

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        if all(e == 2 for e in self._qid_shape):
            return f'cirq.MatrixGate({proper_repr(self._matrix)})'
        return (f'cirq.MatrixGate({proper_repr(self._matrix)}, '
                f'qid_shape={self._qid_shape})')

    def __str__(self):
        return str(self._matrix.round(3))


class SingleQubitMatrixGate(MatrixGate, gate_features.SingleQubitGate):
    """A 1-qubit or qudit gate defined by its matrix.

    More general than specialized classes like `ZPowGate`, but more expensive
    and more float-error sensitive to work with (due to using
    eigendecompositions).
    """

    @deprecated(deadline='v0.8',
                fix='Use `cirq.MatrixGate` instead.',
                func_name='cirq.SingleQubitMatrixGate')
    def __init__(self, matrix: np.ndarray) -> None:
        """
        Initializes the single qubit matrix gate.

        Args:
            matrix: The matrix that defines the gate.
        """
        super().__init__(matrix, qid_shape=(matrix.shape[0],))

    def __repr__(self):
        return 'cirq.SingleQubitMatrixGate({})'.format(proper_repr(
            self._matrix))

    def _json_dict_(self):
        return {
            'cirq_type': self.__class__.__name__,
            'matrix': self._matrix,
        }

    @classmethod
    def _from_json_dict_(cls, matrix, **kwargs):
        return cls(matrix=np.array(matrix))


class TwoQubitMatrixGate(MatrixGate, gate_features.TwoQubitGate):
    """A 2-qubit gate defined only by its matrix.

    More general than specialized classes like `CZPowGate`, but more expensive
    and more float-error sensitive to work with (due to using
    eigendecompositions).
    """

    @deprecated(deadline='v0.8',
                fix='Use `cirq.MatrixGate` instead.',
                func_name='cirq.TwoQubitMatrixGate')
    def __init__(self, matrix: np.ndarray) -> None:
        """
        Initializes the 2-qubit matrix gate.

        Args:
            matrix: The matrix that defines the gate.
        """
        super().__init__(matrix, qid_shape=(2, 2))

    def _json_dict_(self):
        return {
            'cirq_type': self.__class__.__name__,
            'matrix': self._matrix,
        }

    @classmethod
    def _from_json_dict_(cls, matrix, **kwargs):
        return cls(matrix=np.array(matrix))

    def __repr__(self):
        return 'cirq.TwoQubitMatrixGate({})'.format(proper_repr(self._matrix))


def _matrix_to_diagram_symbol(matrix: np.ndarray,
                              args: 'protocols.CircuitDiagramInfoArgs') -> str:
    if args.precision is not None:
        matrix = matrix.round(args.precision)
    result = str(matrix)
    if args.use_unicode_characters:
        lines = result.split('\n')
        for i in range(len(lines)):
            lines[i] = lines[i].replace('[[', '')
            lines[i] = lines[i].replace(' [', '')
            lines[i] = lines[i].replace(']', '')
        w = max(len(line) for line in lines)
        for i in range(len(lines)):
            lines[i] = '│' + lines[i].ljust(w) + '│'
        lines.insert(0, '┌' + ' ' * w + '┐')
        lines.append('└' + ' ' * w + '┘')
        result = '\n'.join(lines)
    return result

==> cirq_unstable-0.8.0.dev20200129194008-py3-none-any.whl: cirq/ops/matrix_gates.py <==
# Copyright 2018 The Cirq Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Quantum gates defined by a matrix."""

from typing import cast, Any, Tuple, Optional, Iterable, TYPE_CHECKING

import numpy as np

from cirq import linalg, protocols
from cirq._compat import proper_repr, deprecated
from cirq.ops import gate_features, raw_types

if TYPE_CHECKING:
    import cirq


class MatrixGate(raw_types.Gate):
    """A unitary qubit or qudit gate defined entirely by its matrix."""

    def __init__(self,
                 matrix: np.ndarray,
                 *,
                 qid_shape: Optional[Iterable[int]] = None) -> None:
        """Initializes a matrix gate.
        Args:
            matrix: The matrix that defines the gate.
            qid_shape: The shape of state tensor that the matrix applies to.
                If not specified, this value is inferred by assuming that the
                matrix is supposed to apply to qubits.
        """
        if len(matrix.shape) != 2 or matrix.shape[0] != matrix.shape[1]:
            raise ValueError('`matrix` must be a square 2d numpy array.')

        if qid_shape is None:
            n = int(np.round(np.log2(matrix.shape[0] or 1)))
            if 2**n != matrix.shape[0]:
                raise ValueError(
                    f'Matrix width ({matrix.shape[0]}) is not a power of 2 and '
                    f'qid_shape is not specified.')
            qid_shape = (2,) * n

        self._matrix = matrix
        self._qid_shape = tuple(qid_shape)
        m = int(np.prod(self._qid_shape))
        if self._matrix.shape != (m, m):
            raise ValueError('Wrong matrix shape for qid_shape.\n'
                             f'Matrix shape: {self._matrix.shape}\n'
                             f'qid_shape: {self._qid_shape}\n')

        if not linalg.is_unitary(matrix):
            raise ValueError(f'Not a unitary matrix: {self._matrix}')

    def _json_dict_(self):
        return {
            'cirq_type': self.__class__.__name__,
            'matrix': self._matrix.tolist(),
            'qid_shape': self._qid_shape,
        }

    @classmethod
    def _from_json_dict_(cls, matrix, qid_shape, **kwargs):
        return cls(matrix=np.array(matrix), qid_shape=qid_shape)

    def _qid_shape_(self) -> Tuple[int, ...]:
        return self._qid_shape

    def __pow__(self, exponent: Any) -> 'MatrixGate':
        if not isinstance(exponent, (int, float)):
            return NotImplemented
        e = cast(float, exponent)
        new_mat = linalg.map_eigenvalues(self._matrix, lambda b: b**e)
        return MatrixGate(new_mat, qid_shape=self._qid_shape)

    def _phase_by_(self, phase_turns: float, qubit_index: int) -> 'MatrixGate':
        if not isinstance(phase_turns, (int, float)):
            return NotImplemented
        if self._qid_shape[qubit_index] != 2:
            return NotImplemented
        result = np.copy(self._matrix).reshape(self._qid_shape * 2)

        p = np.exp(2j * np.pi * phase_turns)
        i = qubit_index
        j = qubit_index + len(self._qid_shape)
        result[linalg.slice_for_qubits_equal_to([i], 1)] *= p
        result[linalg.slice_for_qubits_equal_to([j], 1)] *= np.conj(p)
        return MatrixGate(matrix=result.reshape(self._matrix.shape),
                          qid_shape=self._qid_shape)

    def _has_unitary_(self) -> bool:
        return True

    def _unitary_(self) -> np.ndarray:
        return np.copy(self._matrix)

    def _circuit_diagram_info_(self, args: 'cirq.CircuitDiagramInfoArgs'
                              ) -> 'cirq.CircuitDiagramInfo':
        main = _matrix_to_diagram_symbol(self._matrix, args)
        rest = [f'#{i+1}' for i in range(1, len(self._qid_shape))]
        return protocols.CircuitDiagramInfo(wire_symbols=[main, *rest])

    def __hash__(self):
        vals = tuple(v for _, v in np.ndenumerate(self._matrix))
        return hash((MatrixGate, vals))

    def _approx_eq_(self, other: Any, atol) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return np.allclose(self._matrix, other._matrix, rtol=0, atol=atol)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return (self._qid_shape == other._qid_shape and
                np.array_equal(self._matrix, other._matrix))

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        if all(e == 2 for e in self._qid_shape):
            return f'cirq.MatrixGate({proper_repr(self._matrix)})'
        return (f'cirq.MatrixGate({proper_repr(self._matrix)}, '
                f'qid_shape={self._qid_shape})')

    def __str__(self):
        return str(self._matrix.round(3))


class SingleQubitMatrixGate(MatrixGate, gate_features.SingleQubitGate):
    """A 1-qubit or qudit gate defined by its matrix.

    More general than specialized classes like `ZPowGate`, but more expensive
    and more float-error sensitive to work with (due to using
    eigendecompositions).
    """

    @deprecated(deadline='v0.8',
                fix='Use `cirq.MatrixGate` instead.',
                name='cirq.SingleQubitMatrixGate')
    def __init__(self, matrix: np.ndarray) -> None:
        """
        Initializes the single qubit matrix gate.

        Args:
            matrix: The matrix that defines the gate.
        """
        super().__init__(matrix, qid_shape=(matrix.shape[0],))

    def __repr__(self):
        return 'cirq.SingleQubitMatrixGate({})'.format(proper_repr(
            self._matrix))

    def _json_dict_(self):
        return {
            'cirq_type': self.__class__.__name__,
            'matrix': self._matrix,
        }

    @classmethod
    def _from_json_dict_(cls, matrix, **kwargs):
        return cls(matrix=np.array(matrix))


class TwoQubitMatrixGate(MatrixGate, gate_features.TwoQubitGate):
    """A 2-qubit gate defined only by its matrix.

    More general than specialized classes like `CZPowGate`, but more expensive
    and more float-error sensitive to work with (due to using
    eigendecompositions).
    """

    @deprecated(deadline='v0.8',
                fix='Use `cirq.MatrixGate` instead.',
                name='cirq.TwoQubitMatrixGate')
    def __init__(self, matrix: np.ndarray) -> None:
        """
        Initializes the 2-qubit matrix gate.

        Args:
            matrix: The matrix that defines the gate.
        """
        super().__init__(matrix, qid_shape=(2, 2))

    def _json_dict_(self):
        return {
            'cirq_type': self.__class__.__name__,
            'matrix': self._matrix,
        }

    @classmethod
    def _from_json_dict_(cls, matrix, **kwargs):
        return cls(matrix=np.array(matrix))

    def __repr__(self):
        return 'cirq.TwoQubitMatrixGate({})'.format(proper_repr(self._matrix))


def _matrix_to_diagram_symbol(matrix: np.ndarray,
                              args: 'protocols.CircuitDiagramInfoArgs') -> str:
    if args.precision is not None:
        matrix = matrix.round(args.precision)
    result = str(matrix)
    if args.use_unicode_characters:
        lines = result.split('\n')
        for i in range(len(lines)):
            lines[i] = lines[i].replace('[[', '')
            lines[i] = lines[i].replace(' [', '')
            lines[i] = lines[i].replace(']', '')
        w = max(len(line) for line in lines)
        for i in range(len(lines)):
            lines[i] = '│' + lines[i].ljust(w) + '│'
        lines.insert(0, '┌' + ' ' * w + '┐')
        lines.append('└' + ' ' * w + '┘')
        result = '\n'.join(lines)
    return result
